from typing import T
from typing import Union
from xml.etree import ElementTree

from defusedxml.ElementTree import fromstring


class XMLParseError(Exception):
    pass


def serialize_xml(write, elem: ElementTree.Element, qnames, namespaces, indent="\n"):
    tag = elem.tag
    text = elem.text
    if text:
        text = text.strip()
    if tag is ElementTree.Comment:
        write(f"{indent}<!--{text}-->")
    elif tag is ElementTree.ProcessingInstruction:
        write(f"{indent}<?{text}?>")
    else:
        tag = qnames[tag]
        next_indent = f"{indent}  "
        if tag is None:
            if text:
                write(indent)
                write(ElementTree._escape_cdata(text))
            for e in elem:
                serialize_xml(write, e, qnames, None, next_indent)
        else:
            write(f"{indent}<{tag}")
            items = list(elem.items())
            if items or namespaces:
                if namespaces:
                    for v, k in sorted(namespaces.items(), key=lambda x: x[1]):  # sort on prefix
                        if k:
                            k = ":" + k
                        write(f' xmlns{k}="{ElementTree._escape_attrib(v)}"')
                for k, v in items:
                    if isinstance(k, ElementTree.QName):
                        k = k.text
                    v = qnames[v.text] if isinstance(v, ElementTree.QName) else ElementTree._escape_attrib(v)
                    write(f' {qnames[k]}="{v}"')
            children = len(elem)
            if text or children:
                write(">")
                if text:
                    if children:
                        write(next_indent)
                    write(ElementTree._escape_cdata(text))
                for e in elem:
                    serialize_xml(write, e, qnames, None, next_indent)
                if children:
                    write(indent)
                if text:
                    write(f"</{tag}>")
                else:
                    write(f"</{tag}>")
            else:
                write(" />")

    tail = elem.tail
    if tail:
        tail = tail.strip()
    if tail:
        write(indent)
        write(ElementTree._escape_cdata(tail))


class XML:
    element: ElementTree.Element

    def __init__(self, element, namespaces=None):
        self.element = element
        self.namespaces = namespaces or {}

    @classmethod
    def fromstring(cls, xml, namespaces=None):
        try:
            return cls(
                fromstring(xml, forbid_dtd=True, forbid_entities=True, forbid_external=True),
                namespaces=namespaces,
            )
        except Exception as exc:
            raise XMLParseError(exc) from exc

    def __repr__(self):
        return ElementTree.tostring(self.element, encoding="unicode")

    def __str__(self):
        output = []
        serialize_xml(output.append, self.element, *ElementTree._namespaces(self.element))
        return "".join(output).strip()

    def find(self, tag, *, strict=True) -> Union["XML", None]:
        # noinspection StrFormat
        tag = tag.format(**self.namespaces)
        element = self.element if self.element.tag == tag else self.element.find(tag)

        if element is None:
            if strict:
                children = ",".join(f"<{element.tag}>" for element in self.element)
                raise XMLParseError(f"Failed to find <{tag}> element. Element list was: {children}. Parent: {self.element}")
        else:
            return self.__class__(element, self.namespaces)

    def findall(self, tag, *, strict=True) -> list["XML"]:
        # noinspection StrFormat
        tag = tag.format(**self.namespaces)
        if self.element.tag == tag:
            return [self]
        else:
            elements = self.element.findall(tag)

        return [self.__class__(element, self.namespaces) for element in elements]

    def text(self, tag=None, strict=True, cast: type[T] | None = None) -> str | T:
        value = None
        if tag:
            xml = self.find(tag, strict=strict)
            if xml:
                value = xml.text()
        elif self.element.text:
            value = self.element.text.strip()

        if cast is str:
            return value or ""
        elif cast and (value or strict):
            try:
                return cast(value)
            except Exception as exc:
                raise XMLParseError(f"Failed casting <{tag}>{value!r} to {cast}", exc) from exc
        else:
            return value

    def attr(self, attr, strict=True, cast: type[T] | None = None) -> str | T:
        # noinspection StrFormat
        attr = attr.format(**self.namespaces)
        attrs = self.element.attrib
        if attr in attrs:
            value = attrs[attr]
        elif strict:
            raise XMLParseError(f"Failed to find attribute {attr!r} in element {self.element.tag} with attributes {attrs}.")
        else:
            value = None

        if cast is str:
            return value or ""
        elif cast and (value or strict):
            try:
                return cast(value)
            except Exception as exc:
                raise XMLParseError(f"Failed casting {attr}={value!r} to {cast}", exc) from exc
        else:
            return value
